@@ -22,10 +22,15 @@ import java.lang.reflect.Field;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
+import java.util.Collections;
 import java.util.Hashtable;
 import java.util.List;
+import java.util.Observable;
+import java.util.Observer;
 import java.util.concurrent.ConcurrentMap;
+import java.util.concurrent.CountDownLatch;
 import java.util.concurrent.atomic.AtomicBoolean;
+import java.util.concurrent.atomic.AtomicInteger;
 
 import junit.framework.TestCase;
 
@@ -65,6 +70,7 @@ public class ServiceRegistryTest extends TestCase
         ServiceRegistry sr = new ServiceRegistry(new Logger(), null);
         EventHook hook = new EventHook()
         {
+            @Override
             public void event(ServiceEvent event, Collection contexts)
             {
             }
@@ -133,6 +139,7 @@ public class ServiceRegistryTest extends TestCase
         ServiceRegistry sr = new ServiceRegistry(new Logger(), null);
         FindHook hook = new FindHook()
         {
+            @Override
             public void find(BundleContext context, String name, String filter,
                 boolean allServices, Collection references)
             {
@@ -201,10 +208,12 @@ public class ServiceRegistryTest extends TestCase
         ServiceRegistry sr = new ServiceRegistry(new Logger(), null);
         ListenerHook hook = new ListenerHook()
         {
+            @Override
             public void added(Collection listeners)
             {
             }
 
+            @Override
             public void removed(Collection listener)
             {
             }
@@ -272,23 +281,28 @@ public class ServiceRegistryTest extends TestCase
         ServiceRegistry sr = new ServiceRegistry(new Logger(), null);
         class CombinedService implements ListenerHook, FindHook, EventHook, Runnable
         {
+            @Override
             public void added(Collection listeners)
             {
             }
 
+            @Override
             public void removed(Collection listener)
             {
             }
 
+            @Override
             public void find(BundleContext context, String name, String filter,
                     boolean allServices, Collection references)
             {
             }
 
+            @Override
             public void event(ServiceEvent event, Collection contexts)
             {
             }
 
+            @Override
             public void run()
             {
             }
@@ -385,7 +399,7 @@ public class ServiceRegistryTest extends TestCase
         final ServiceHolder sh = new ServiceHolder();
         uc.m_svcHolderRef.set(sh);
 
-        final StringBuilder sb = new StringBuilder();
+        final StringBuffer sb = new StringBuffer();
         final AtomicBoolean threadException = new AtomicBoolean(false);
         Thread t = new Thread() {
             @Override
@@ -506,9 +520,8 @@ public class ServiceRegistryTest extends TestCase
 
         inUseMap.put(b, new UsageCount[] {uc});
 
-        assertTrue(sr.ungetService(b, ref, null));
+        assertFalse(sr.ungetService(b, ref, null));
         assertNull(uc.m_svcHolderRef.get());
-        assertNull(inUseMap.get(b));
     }
 
     @SuppressWarnings("unchecked")
@@ -538,7 +551,6 @@ public class ServiceRegistryTest extends TestCase
 
         assertTrue(sr.ungetService(b, ref, null));
         assertNull(uc.m_svcHolderRef.get());
-        assertNull(inUseMap.get(b));
 
         Mockito.verify(reg, Mockito.times(1)).
             ungetService(Mockito.isA(Bundle.class), Mockito.eq(svc));
@@ -596,7 +608,6 @@ public class ServiceRegistryTest extends TestCase
 
         assertTrue(sr.ungetService(b, ref, null));
         assertNull(uc.m_svcHolderRef.get());
-        assertNull(inUseMap.get(b));
 
         Mockito.verify(reg, Mockito.never()).
             ungetService(Mockito.isA(Bundle.class), Mockito.any());
@@ -638,7 +649,6 @@ public class ServiceRegistryTest extends TestCase
             assertEquals("Test!", re.getMessage());
         }
         assertNull(uc.m_svcHolderRef.get());
-        assertNull(inUseMap.get(b));
 
         Mockito.verify(reg, Mockito.times(1)).ungetService(b, svc);
     }
@@ -1033,6 +1043,191 @@ public class ServiceRegistryTest extends TestCase
         assertSame(uc2, inUseMap.get(b)[0]);
     }
 
+    public void testGetUngetServiceFactory() throws Exception
+    {
+        final ServiceRegistry sr = new ServiceRegistry(null, null);
+        final Bundle regBundle = Mockito.mock(Bundle.class);
+        final ServiceRegistration<?> reg = sr.registerService(regBundle, new String[] {Observer.class.getName()},
+                new ServiceFactory<Observer>()
+                {
+
+                    final class ObserverImpl implements Observer
+                    {
+                        private final AtomicInteger counter = new AtomicInteger();
+                        public volatile boolean active = true;
+
+                        @Override
+                        public void update(Observable o, Object arg)
+                        {
+                            counter.incrementAndGet();
+                            if ( !active )
+                            {
+                                throw new IllegalArgumentException("Iteration:" + counter.get());
+                            }
+                        }
+
+                    };
+
+                    @Override
+                    public Observer getService(Bundle bundle, ServiceRegistration<Observer> registration)
+                    {
+                        return new ObserverImpl();
+                    }
+
+                    @Override
+                    public void ungetService(Bundle bundle, ServiceRegistration<Observer> registration, Observer service)
+                    {
+                        ((ObserverImpl)service).active = false;
+                    }
+                }, null);
+
+        final Bundle clientBundle = Mockito.mock(Bundle.class);
+        Mockito.when(clientBundle.getBundleId()).thenReturn(42L);
+
+        // check simple get/unget
+        final Object obj = sr.getService(clientBundle, reg.getReference(), false);
+        assertNotNull(obj);
+        assertTrue(obj instanceof Observer);
+        ((Observer)obj).update(null, null);
+        sr.ungetService(clientBundle, reg.getReference(), null);
+        try {
+            ((Observer)obj).update(null, null);
+            fail();
+        }
+        catch ( final IllegalArgumentException iae)
+        {
+            // expected
+        }
+
+        // start three threads
+        final int MAX_THREADS = 3;
+        final int MAX_LOOPS = 50000;
+        final CountDownLatch latch = new CountDownLatch(MAX_THREADS);
+        final Thread[] threads = new Thread[MAX_THREADS];
+        final List<Exception> exceptions = Collections.synchronizedList(new ArrayList<Exception>());
+        for(int i=0; i<MAX_THREADS; i++)
+        {
+            threads[i] = new Thread(new Runnable()
+            {
+
+                @Override
+                public void run()
+                {
+                    try
+                    {
+                        Thread.currentThread().sleep(50);
+                    }
+                    catch (InterruptedException e1)
+                    {
+                        // ignore
+                    }
+                    for(int i=0; i < MAX_LOOPS; i++)
+                    {
+                        try
+                        {
+                            final Object obj = sr.getService(clientBundle, reg.getReference(), false);
+                            ((Observer)obj).update(null, null);
+                            sr.ungetService(clientBundle, reg.getReference(), null);
+                        }
+                        catch ( final Exception e)
+                        {
+                            exceptions.add(e);
+                        }
+                    }
+                    latch.countDown();
+                }
+            });
+        }
+        for(int i=0; i<MAX_THREADS; i++)
+        {
+            threads[i].start();
+        }
+
+        latch.await();
+
+        List<String> counterValues = new ArrayList<String>();
+        for (Exception ex : exceptions)
+        {
+            counterValues.add(ex.getMessage());
+        }
+
+        assertTrue("" + counterValues, exceptions.isEmpty());
+    }
+
+    public void testUsageCountCleanup() throws Exception
+    {
+        ServiceRegistry sr = new ServiceRegistry(null, null);
+        Bundle regBundle = Mockito.mock(Bundle.class);
+
+        ServiceRegistration<?> reg = sr.registerService(
+                regBundle, new String [] {String.class.getName()}, "hi", null);
+
+        final Bundle clientBundle = Mockito.mock(Bundle.class);
+        Mockito.when(clientBundle.getBundleId()).thenReturn(327L);
+
+        assertEquals("hi", sr.getService(clientBundle, reg.getReference(), false));
+        sr.ungetService(clientBundle, reg.getReference(), null);
+
+        ConcurrentMap<Bundle, UsageCount[]> inUseMap =
+                (ConcurrentMap<Bundle, UsageCount[]>) getPrivateField(sr, "m_inUseMap");
+
+        sr.unregisterService(regBundle, reg);
+        assertEquals(0, inUseMap.size());
+    }
+
+    @SuppressWarnings("unchecked")
+    public void testGetServiceThrowsException() throws Exception
+    {
+        final ServiceRegistry sr = new ServiceRegistry(null, null);
+
+        final Bundle b = Mockito.mock(Bundle.class);
+        ServiceRegistrationImpl reg = Mockito.mock(ServiceRegistrationImpl.class);
+        Mockito.when(reg.isValid()).thenReturn(true);
+        Mockito.when(reg.getService(b)).thenAnswer(new Answer() {
+            @Override
+            public Object answer(InvocationOnMock invocation) throws Throwable
+            {
+                Thread.sleep(500);
+                throw new Exception("boo!");
+            }
+        });
+
+        final ServiceReferenceImpl ref = Mockito.mock(ServiceReferenceImpl.class);
+        Mockito.when(ref.getRegistration()).thenReturn(reg);
+
+        final StringBuffer sb = new StringBuffer();
+        Thread t = new Thread()
+        {
+            @Override
+            public void run()
+            {
+                try
+                {
+                    assertEquals("Should not yet have given the service to the other thread",
+                            "", sb.toString());
+                    sr.getService(b, ref, false);
+                }
+                catch (Exception e)
+                {
+                    // We expect an exception here.
+                }
+            }
+        };
+        t.start();
+
+        // Wait until the other thread has called getService();
+        Thread.sleep(250);
+
+        // This thread has waited long enough for the other thread to call getService()
+        // however the actual getService() call blocks long enough for this one to then
+        // concurrently call getService() while the other thread is in getService() of the
+        // factory. This thread will then end up in m_latch.await().
+        // The factory implementation of the other thread then throws an exception. This test
+        // ultimately checks that this thread here is not stuck waiting forwever.
+        assertNull(sr.getService(b, ref, false));
+        sb.append("Obtained service");
+    }
+
     private Object getPrivateField(Object obj, String fieldName) throws NoSuchFieldException,
             IllegalAccessException
     {
